@@ -23,6 +23,8 @@ import weakref
 import typing
 
 import chess
+import chess.engine
+import chess.svg
 
 from typing import Callable, Dict, Generic, Iterable, Iterator, List, Mapping, MutableMapping, Set, TextIO, Tuple, Type, TypeVar, Optional, Union
 
@@ -106,6 +108,24 @@ MOVETEXT_REGEX = re.compile(r"""
 
 SKIP_MOVETEXT_REGEX = re.compile(r""";|\{|\}""")
 
+
+CLOCK_REGEX = re.compile(r"""\[%clk\s(\d+):(\d+):(\d+)\]""")
+
+EVAL_REGEX = re.compile(r"""
+    \[%eval\s(?:
+        \#([+-]?\d+)
+        |([+-]?(?:\d{0,10}\.\d{1,2}|\d{1,10}\.?))
+    )\]
+    """, re.VERBOSE)
+
+ARROWS_REGEX = re.compile(r"""
+    \[%(?:csl|cal)\s(
+        [RGYB][a-h][1-8](?:[a-h][1-8])?
+        (?:,[RGYB][a-h][1-8](?:[a-h][1-8])?)*
+    )\]
+    """, re.VERBOSE)
+
+
 TAG_ROSTER = ["Event", "Site", "Date", "Round", "White", "Black", "Result"]
 
 
@@ -314,7 +334,7 @@ class GameNode:
         """Removes a variation."""
         self.variations.remove(self.variation(move))
 
-    def add_variation(self, move: chess.Move, *, comment: str = "", starting_comment: str = "", nags: Iterable[int] = ()) -> "GameNode":
+    def add_variation(self, move: chess.Move, *, comment: str = "", starting_comment: str = "", nags: Iterable[int] = []) -> "GameNode":
         """Creates a child node with the given attributes."""
         node = type(self).dangling_node()
         node.move = move
@@ -344,7 +364,7 @@ class GameNode:
         """Returns an iterator over the main moves after this node."""
         return Mainline(self, lambda node: node._move())
 
-    def add_line(self, moves: Iterable[chess.Move], *, comment: str = "", starting_comment: str = "", nags: Iterable[int] = ()) -> "GameNode":
+    def add_line(self, moves: Iterable[chess.Move], *, comment: str = "", starting_comment: str = "", nags: Iterable[int] = []) -> "GameNode":
         """
         Creates a sequence of child nodes for the given list of moves.
         Adds *comment* and *nags* to the last node of the line and returns it.
@@ -366,6 +386,138 @@ class GameNode:
 
         return node
 
+    def eval(self) -> Optional[chess.engine.PovScore]:
+        """
+        Parses the first valid ``[%eval ...]`` annotation in the comment of
+        this node, if any.
+        """
+        match = EVAL_REGEX.search(self.comment)
+        if not match:
+            return None
+        if match.group(1):
+            mate = int(match.group(1))
+            return chess.engine.PovScore(chess.engine.Mate(mate), chess.WHITE) if mate else None
+        else:
+            return chess.engine.PovScore(chess.engine.Cp(int(float(match.group(2)) * 100)), chess.WHITE)
+
+    def set_eval(self, score: Optional[chess.engine.PovScore]) -> None:
+        """
+        Replaces the first valid ``[%eval ...]`` annotation in the comment of
+        this node or adds a new one.
+        """
+        eval = ""
+        if score is not None:
+            cp = score.white().score()
+            if cp is not None:
+                eval = f"[%eval {float(cp) / 100:.2f}]"
+            elif score.white().mate():
+                eval = f"[%eval #{score.white().mate()}]"
+
+        self.comment, found = EVAL_REGEX.subn(eval, self.comment, count=1)
+
+        if not found and eval:
+            if not self.comment.endswith(" "):
+                self.comment += " "
+            self.comment += eval
+
+    def arrows(self) -> List[chess.svg.Arrow]:
+        """
+        Parses all ``[%csl ...]`` and ``[%cal ...]`` annotations in the comment
+        of this node.
+
+        Returns a list of :class:`arrows <chess.svg.Arrow>`.
+        """
+        arrows = []
+        for match in ARROWS_REGEX.finditer(self.comment):
+            for group in match.group(1).split(","):
+                color = "G"
+                if group[0] == "R":
+                    color = "red"
+                elif group[0] == "Y":
+                    color = "yellow"
+                elif group[0] == "B":
+                    color = "blue"
+
+                tail = chess.SQUARE_NAMES.index(group[1:3])
+                if len(group) == 5:
+                    head = chess.SQUARE_NAMES.index(group[3:5])
+                else:
+                    head = tail
+
+                arrows.append(chess.svg.Arrow(tail, head, color=color))
+
+        return arrows
+
+    def set_arrows(self, arrows: Iterable[Union[chess.svg.Arrow, Tuple[chess.Square, chess.Square]]]) -> None:
+        """
+        Replaces all valid ``[%csl ...]`` and ``[%cal ...]`` annotations in
+        the comment of this node or adds new ones.
+        """
+        csl = []
+        cal = []
+
+        for arrow in arrows:
+            try:
+                tail, head, color = arrow.tail, arrow.head, arrow.color  # type: ignore
+                if color == "red":
+                    color = "R"
+                elif color == "yellow":
+                    color = "Y"
+                elif color == "blue":
+                    color = "B"
+                else:
+                    color = "G"
+            except AttributeError:
+                tail, head = arrow  # type: ignore
+                color = "G"
+            if tail == head:
+                csl.append(f"{color}{chess.SQUARE_NAMES[tail]}")
+            else:
+                cal.append(f"{color}{chess.SQUARE_NAMES[tail]}{chess.SQUARE_NAMES[head]}")
+
+        self.comment = ARROWS_REGEX.sub("", self.comment).strip()
+
+        prefix = ""
+        if csl:
+            prefix += f"[%csl {','.join(csl)}]"
+        if cal:
+            prefix += f"[%cal {','.join(cal)}]"
+
+        if prefix:
+            self.comment = prefix + " " + self.comment
+
+    def clock(self) -> Optional[float]:
+        """
+        Parses the first valid ``[%clk ...]`` annotation in the comment of
+        this node, if any.
+
+        Returns the player's remaining time to the next time control after this
+        move, in seconds.
+        """
+        match = CLOCK_REGEX.search(self.comment)
+        if match is None:
+            return None
+        return int(match.group(1)) * 3600 + int(match.group(2)) * 60 + int(match.group(3))
+
+    def set_clock(self, seconds: Optional[float]) -> None:
+        """
+        Replaces the first valid ``[%clk ...]`` annotation in the comment of
+        this node or adds a new one.
+        """
+        clk = ""
+        if seconds is not None:
+            hours = seconds // 3600
+            minutes = seconds % 3600 // 60
+            seconds = seconds % 3600 % 60
+            clk = f"[%clk {hours:d}:{minutes:02d}:{seconds:02d}]"
+
+        self.comment, found = CLOCK_REGEX.subn(clk, self.comment, count=1)
+
+        if not found and clk:
+            if not self.comment.endswith(" "):
+                self.comment += " "
+            self.comment += clk
+
     def _accept_node(self, parent_board: chess.Board, visitor: "BaseVisitor[ResultT]") -> None:
         assert self.move is not None, "cannot visit dangling GameNode"
 
@@ -1117,7 +1269,7 @@ class StringExporter(StringExporterMixin, BaseVisitor[str]):
         return self.result()
 
 
-class FileExporter(StringExporterMixin, BaseVisitor[None]):
+class FileExporter(StringExporterMixin, BaseVisitor[int]):
     """
     Acts like a :class:`~chess.pgn.StringExporter`, but games are written
     directly into a text file.
@@ -1138,19 +1290,23 @@ class FileExporter(StringExporterMixin, BaseVisitor[None]):
         super().__init__(columns=columns, headers=headers, comments=comments, variations=variations)
         self.handle = handle
 
+    def begin_game(self):
+        self.written: int = 0
+        super().begin_game()
+
     def flush_current_line(self) -> None:
         if self.current_line:
-            self.handle.write(self.current_line.rstrip())
-            self.handle.write("\n")
+            self.written += self.handle.write(self.current_line.rstrip())
+            self.written += self.handle.write("\n")
         self.current_line = ""
 
     def write_line(self, line: str = "") -> None:
         self.flush_current_line()
-        self.handle.write(line.rstrip())
-        self.handle.write("\n")
+        self.written += self.handle.write(line.rstrip())
+        self.written += self.handle.write("\n")
 
-    def result(self) -> None:
-        return None
+    def result(self) -> int:
+        return self.written
 
     def __repr__(self) -> str:
         return f"<FileExporter at {id(self):#x}>"
@@ -1226,12 +1382,19 @@ def read_game(handle: TextIO, *, Visitor = GameBuilder):
         line = handle.readline()
 
     # Parse game headers.
+    consecutive_empty_lines = 0
     while line:
         # Ignore comments.
         if line.startswith("%") or line.startswith(";"):
             line = handle.readline()
             continue
 
+        # Ignore up to one consecutive empty line between headers.
+        if consecutive_empty_lines < 1 and line.isspace():
+            consecutive_empty_lines += 1
+            line = handle.readline()
+            continue
+
         # First token of the game.
         if not found_game:
             found_game = True
@@ -1244,6 +1407,8 @@ def read_game(handle: TextIO, *, Visitor = GameBuilder):
         if not line.startswith("["):
             break
 
+        consecutive_empty_lines = 0
+
         if not skipping_game:
             tag_match = TAG_REGEX.match(line)
             if tag_match:
@@ -1261,10 +1426,6 @@ def read_game(handle: TextIO, *, Visitor = GameBuilder):
     if not skipping_game:
         skipping_game = visitor.end_headers() is SKIP
 
-    # Ignore single empty line after headers.
-    if line.isspace():
-        line = handle.readline()
-
     if not skipping_game:
         # Chess variant.
         headers = managed_headers if unmanaged_headers is None else unmanaged_headers
